#include "xine_info.h"

/************************************************************************************************
* xine OSD definitions                                                                           *
************************************************************************************************/

//#ifdef __cplusplus
extern "C"
{
// #endif
#define OVL_PALETTE_SIZE 256

//#ifdef        __GNUC__
//#define CLUT_Y_CR_CB_INIT(_y,_cr,_cb) {y: (_y), cr: (_cr), cb: (_cb)}
//#else
#define CLUT_Y_CR_CB_INIT(_y,_cr,_cb)	{ (_cb), (_cr), (_y) }
//#endif

	typedef struct clut_t
	{			/* CLUT == Color LookUp Table */
		uint8_t cb:8;
		uint8_t cr:8;
		uint8_t y:8;
		uint8_t foo:8;
	}
	__attribute__ ((packed)) clut_t;

	clut_t textpalettes_color[] = {
		/* white, no border, translucid */
		CLUT_Y_CR_CB_INIT (0x00, 0x00, 0x00),	//0
		CLUT_Y_CR_CB_INIT (0x60, 0x80, 0x80),	//1
		CLUT_Y_CR_CB_INIT (0x70, 0x80, 0x80),	//2
		CLUT_Y_CR_CB_INIT (0x80, 0x80, 0x80),	//3
		CLUT_Y_CR_CB_INIT (0x80, 0x80, 0x80),	//4
		CLUT_Y_CR_CB_INIT (0x80, 0x80, 0x80),	//5
		CLUT_Y_CR_CB_INIT (0x80, 0x80, 0x80),	//6
		CLUT_Y_CR_CB_INIT (0xa0, 0x80, 0x80),	//7
		CLUT_Y_CR_CB_INIT (0xc0, 0x80, 0x80),	//8
		CLUT_Y_CR_CB_INIT (0xe0, 0x80, 0x80),	//9
		CLUT_Y_CR_CB_INIT (0xff, 0x80, 0x80),	//10
		/* yellow, black border, transparent */
		CLUT_Y_CR_CB_INIT (0x00, 0x00, 0x00),	//0
		CLUT_Y_CR_CB_INIT (0x80, 0x80, 0xe0),	//1
		CLUT_Y_CR_CB_INIT (0x80, 0x80, 0xc0),	//2
		CLUT_Y_CR_CB_INIT (0x60, 0x80, 0xa0),	//3
		CLUT_Y_CR_CB_INIT (0x40, 0x80, 0x80),	//4
		CLUT_Y_CR_CB_INIT (0x20, 0x80, 0x80),	//5
		CLUT_Y_CR_CB_INIT (0x00, 0x80, 0x80),	//6
		CLUT_Y_CR_CB_INIT (0x40, 0x84, 0x60),	//7
		CLUT_Y_CR_CB_INIT (0xd0, 0x88, 0x40),	//8
		CLUT_Y_CR_CB_INIT (0xe0, 0x8a, 0x00),	//9
		CLUT_Y_CR_CB_INIT (0xff, 0x90, 0x00),	//10
	};

	static uint8_t textpalettes_trans[] = {
		/* white, no border, translucid */
		0, 8, 9, 10, 11, 12, 13, 14, 15, 15, 15,
		/* yellow, black border, transparent */
		0, 8, 9, 10, 11, 12, 13, 14, 15, 15, 15,
	};
}				//end extern "C"

static struct xine_status_s
{
	char *symbol;
	int status;
}
xine_status[] =
{
	{
	"�", XINE_STATUS_IDLE}
	,
	{
	"}", XINE_STATUS_STOP}
	,
	{
	">", XINE_STATUS_PLAY}
	,
	{
	"{", XINE_STATUS_QUIT}
	,
	{
	NULL, 0}
};

static struct xine_speeds_s
{
	char *symbol;
	int speed;
}
xine_speeds[] =
{
	{
	"<", XINE_SPEED_PAUSE}
	,
	{
	"@@>", XINE_SPEED_SLOW_4}
	,
	{
	"@>", XINE_SPEED_SLOW_2}
	,
	{
	">", XINE_SPEED_NORMAL}
	,
	{
	">$", XINE_SPEED_FAST_2}
	,
	{
	">$$$", XINE_SPEED_FAST_4}
	,
	{
	NULL, 0}
};

//static uint32_t color[OVL_PALETTE_SIZE];
//static uint8_t trans[OVL_PALETTE_SIZE];

#define BAR_WIDTH 336
#define BAR_HEIGHT 25
#define OSD_BAR_PROGRESS 1
#define OSD_BAR_POS      2
#define OSD_BAR_STEPPER  3


static char *
_osd_get_speed_sym (int speed)
{
	int i;

	for (i = 0; xine_speeds[i].symbol != NULL; i++)
	{
		if (speed == xine_speeds[i].speed)
			return xine_speeds[i].symbol;
	}

	return NULL;
}
static char *
_osd_get_status_sym (int status)
{
	int i;

	for (i = 0; xine_status[i].symbol != NULL; i++)
	{
		if (status == xine_status[i].status)
			return xine_status[i].symbol;
	}

	return NULL;
}

/*This thread is used to count down the osd display, and when timed-out hide it*/
static void *
osd_timer_thread (void *xine_gen)
{
	xine_interface *foo;
	xine_info *xine_data = (xine_info *) xine_gen;
	while (1)
	{
		/*We only need to update once per second */
		sleep (1);
		if (xine_data->osd_info_visible)
		{
			xine_data->osd_info_visible--;
			if (!xine_data->osd_info_visible)
			{
				xine_osd_hide (xine_data->osd_info, 0);
			}
		}
		if (xine_data->osd_bar_visible)
		{
			xine_data->osd_bar_visible--;
			if (!xine_data->osd_bar_visible)
			{
				xine_osd_hide (xine_data->osd_bar[0], 0);
				xine_osd_hide (xine_data->osd_bar[1], 0);
			}
		}
		if (xine_data->osd_status_visible)
		{
			xine_data->osd_status_visible--;
			if (!xine_data->osd_status_visible)
			{
				xine_osd_hide (xine_data->osd_status, 0);
			}
		}
	}
}

/************************************************************************************************
* End Xine OSD definitions                                                                      *
************************************************************************************************/

