
#include "xine_interface.h"
#include <xine.h>

#include <stdlib.h>
#include <sys/time.h>
#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xresource.h>
#include <X11/extensions/XShm.h>

#include "xine_info.h"
#include "osd_defs.h"

#define XINE_ENABLE_EXPERIMENTAL_FEATURES
/* missing stuff from X includes for completion event handler*/
#ifndef XShmGetEventBase
extern int XShmGetEventBase (Display *);
#endif

static void
frame_output_cb (void *xine_gen, int video_width, int video_height,
		 double video_pixel_aspect, int *dest_x, int *dest_y,
		 int *dest_width, int *dest_height, double *dest_pixel_aspect,
		 int *win_x, int *win_y)
{
	xine_info *xine_data = (xine_info *) xine_gen;
	*dest_x = xine_data->video_attr[2];
	*dest_y = xine_data->video_attr[3];
	*win_x = xine_data->video_attr[0];
	*win_y = xine_data->video_attr[1];
	xine_data->stream_w = *dest_width;
	*dest_width = xine_data->video_attr[4];
	xine_data->stream_h = *dest_height;
	*dest_height = xine_data->video_attr[5];

	*dest_pixel_aspect = xine_data->aspect_ratio;
}

static void
dest_size_cb (void *xine_gen, int video_width, int video_height,
	      double video_pixel_aspect, int *dest_width, int *dest_height,
	      double *dest_pixel_aspect)
{
	xine_info *xine_data = (xine_info *) xine_gen;
	/* correct size with video_pixel_aspect */
	xine_data->stream_w = *dest_width;
	*dest_width = xine_data->video_attr[4];
	xine_data->stream_h = *dest_height;
	*dest_height = xine_data->video_attr[5];
	*dest_pixel_aspect = xine_data->aspect_ratio;
}

static void
xine_event_cb (void *xine_iface_gen, const xine_event_t * new_event)
{
	xine_interface *xine_iface = (xine_interface *) xine_iface_gen;

	//fprintf (stderr, "Event recieved with event number: %i \n",
	//       new_event->type);      
	if ((new_event->type == 1) ||
	    (new_event->type == 3) || (new_event->type == 4))
	{
		//xine_iface->add_event(new_event->type, new_event->data, new_event->data_length);
		//char *event_data;             
		if (new_event->data_length)
		{
			xine_ui_data_t *ui_data =
				(xine_ui_data_t *) new_event->data;
			char *event_str;
			event_str = new char[ui_data->str_len + 1];
			strcpy (event_str, ui_data->str);
			//fprintf(stderr,"New ui data of type %i, size %i is: %s\n",new_event->type,ui_data->str_len,ui_data->str);
			xine_iface->set_ui_event (new_event->type, event_str);
		}
		else
			xine_iface->set_ui_event (new_event->type,
						  "nuttin here, please move along");
	}
	else if (new_event->type == 4)
	{
		//int **stream_size;
		//stream_size = (int **)new_event->data;
		//fprintf(stderr,"New video size changed to: %ix%i\n",stream_size[0], stream_size[1]);
	}
}

void
xine_interface::send_simple_event (int new_event)
{
	xine_event_t event;
	event.stream = xine_data->xine_stream;
	event.data = NULL;
	event.data_length = 0;
	event.type = new_event;
	xine_event_send (xine_data->xine_stream, &event);
}

int
xine_interface::send_mouse_event (int x, int y, int button)
{
	x11_rectangle_t rect;
	rect.x = x;
	rect.y = y;
	rect.w = rect.h = 0;
	if (xine_gui_send_vo_data
	    (xine_data->xine_stream, XINE_GUI_SEND_TRANSLATE_GUI_TO_VIDEO,
	     (void *) &rect) == -1)
		return -1;
	int trans_x = rect.x;
	int trans_y = rect.y;

	xine_event_t event;
	event.stream = xine_data->xine_stream;
	xine_input_data_t mouse_data;

	if (button == 0)
	{
		mouse_data.x = trans_x;
		mouse_data.y = trans_y;
		event.type = XINE_EVENT_INPUT_MOUSE_MOVE;
		mouse_data.button = 0;
	}
	else
	{
		event.type = XINE_EVENT_INPUT_MOUSE_BUTTON;
		mouse_data.button = button;
	}

	event.data = &mouse_data;
	event.data_length = sizeof (mouse_data);
	xine_event_send (xine_data->xine_stream, &event);
}

xine_interface::xine_interface ()
{
	xine_data = new xine_info;
	xine_data->xine = xine_new ();
	xine_data->vis_name = "";
	//xine_data->audio_driver = xine_data->video_driver=NULL;

	char home_dir[2048];
	sprintf (home_dir, "%s/.xine/config2", getenv ("HOME"));
	config_file = new char[strlen (home_dir) + 1];
	strcpy (config_file, home_dir);
	xine_config_load (xine_data->xine, config_file);
	xine_init (xine_data->xine);

	const char *const *autoscan_plugins =
		xine_get_autoplay_input_plugin_ids (xine_data->xine);
	char **autoplay_mrls;
	int num_mrls;
	num_plugins = -1;
	fprintf (stderr, "---Autoscan plugins found:\n");
	for (int i = 0; autoscan_plugins[i] != NULL; ++i)
	{
		fprintf (stderr, "     %s -- %s\n",
			 autoscan_plugins[i],
			 xine_get_input_plugin_description (xine_data->xine,
							    autoscan_plugins
							    [i]));
		(const char *const) auto_plugins[i].plugin_id =
			autoscan_plugins[i];
		(const char *) auto_plugins[i].plugin_desc =
			xine_get_input_plugin_description (xine_data->xine,
							   autoscan_plugins
							   [i]);
		num_plugins++;
	}
	const char *const *tmp_drivers;
	tmp_drivers = xine_list_audio_output_plugins (xine_data->xine);
	fprintf (stderr, "---Audio drivers found:\n");
	int i;
	for (i = 0; tmp_drivers[i] != NULL; i++)
		fprintf (stderr, "     %s\n", tmp_drivers[i]);

	tmp_drivers = xine_list_video_output_plugins (xine_data->xine);
	fprintf (stderr, "---Video drivers found:\n");
	for (i = 0; tmp_drivers[i] != NULL; i++)
		fprintf (stderr, "     %s\n", tmp_drivers[i]);

	tmp_drivers = xine_list_post_plugins (xine_data->xine);
	fprintf (stderr, "---Post plugins found:\n");
	for (i = 0; tmp_drivers[i] != NULL; i++)
		fprintf (stderr, "     %s\n", tmp_drivers[i]);

	initialized = false;
}

xine_interface::~xine_interface ()
{
	xine_exit (xine_data->xine);
}

int
xine_interface::get_completion_event_id ()
{
	return xine_data->completion_event;
}

int
xine_interface::send_vo_event (int event_type, void *xine_event)
{
	if (event_type == XINE_GUI_SEND_COMPLETION_EVENT)
	{
		XEvent *comp_event = (XEvent *) xine_event;
		xine_gui_send_vo_data (xine_data->xine_stream, event_type,
				       comp_event);
		return 1;
	}
	else if (event_type == XINE_GUI_SEND_VIDEOWIN_VISIBLE)
	{
		int *is_visible = (int *) xine_event;
		xine_gui_send_vo_data (xine_data->xine_stream, event_type,
				       is_visible);
		return 1;
	}
	else if(event_type == XINE_GUI_SEND_DRAWABLE_CHANGED)
	{
		xine_gui_send_vo_data(xine_data->xine_stream,event_type, xine_event);
		return 1;
	}
	return 0;

}

void
xine_interface::set_parent (Window parent_win, Display * parent_disp,
			    int parent_screen)
{
	double res_h, res_v;
	parent_window = parent_win;
	parent_display = parent_disp;
	parent_screen = parent_screen;
	res_h = (DisplayWidth (parent_display, parent_screen) * 1000
		 / DisplayWidthMM (parent_display, parent_screen));
	res_v = (DisplayHeight (parent_display, parent_screen) * 1000
		 / DisplayHeightMM (parent_display, parent_screen));
	xine_data->aspect_ratio = res_h / res_v;
	xine_data->vis.dest_size_cb = dest_size_cb;
	xine_data->vis.frame_output_cb = frame_output_cb;
	xine_data->vis.user_data = xine_data;
	xine_data->vis.d = parent_window;
	xine_data->vis.display = parent_display;
	xine_data->vis.screen = parent_screen;
}

void
xine_interface::set_video_attr (int win_x, int win_y, int vid_x, int vid_y,
				int *vid_w, int *vid_h)
{
	xine_data->video_attr[0] = win_x;
	xine_data->video_attr[1] = win_y;
	xine_data->video_attr[2] = vid_x;
	xine_data->video_attr[3] = vid_y;
	xine_data->video_attr[4] = *vid_w;
	xine_data->video_attr[5] = *vid_h;
	*vid_w = xine_data->stream_w;
	*vid_h = xine_data->stream_h;
}

bool
xine_interface::init (const char *vid_driver, const char *aud_driver)
{

	/*Detect completion event number, this will have to be sent back through the video driver upon**
	 **each completion event (XEvent with this number) so that the next frame may be displayed     */
	if (XShmQueryExtension ((Display *) xine_data->vis.display) == true)
	{
		xine_data->completion_event =
			XShmGetEventBase ((Display *) xine_data->vis.
					  display) + ShmCompletion;
		fprintf (stderr, "Completion event id: %i\n",
			 xine_data->completion_event);
	}
	else
	{
		xine_data->completion_event = -1;
	}
	if (vid_driver == "auto")
		vid_driver = NULL;
	if (aud_driver == "auto")
		aud_driver = NULL;
	
	xine_data->vis_active = 0;

	/*Initialize video and audio driver */
	xine_data->xine_audio =
		xine_open_audio_driver (xine_data->xine, aud_driver, NULL);

	xine_data->audio_driver = (char *) aud_driver;

	int video_type;
	if (!strcmp (vid_driver, "aa"))
		video_type = XINE_VISUAL_TYPE_AA;
	else if (!strcmp (vid_driver, "SyncFB"))
		video_type = XINE_VISUAL_TYPE_FB;
	else if (!strcmp (vid_driver, "none"))
		video_type = XINE_VISUAL_TYPE_NONE;
	else
		video_type = XINE_VISUAL_TYPE_X11;
	fprintf (stderr, "Attempting to load %s video driver, type %i\n",
		 vid_driver, video_type);
	xine_data->xine_video =
		xine_open_video_driver (xine_data->xine, vid_driver,
					video_type, (void *) &xine_data->vis);

	xine_data->video_driver = (char *) vid_driver;


	xine_data->xine_stream =
		xine_stream_new (xine_data->xine, xine_data->xine_audio,
				 xine_data->xine_video);
	/*Startup a thread to listen and respond to any xine events that are recieved */
	xine_data->xine_event_queue =
		xine_event_new_queue (xine_data->xine_stream);
	xine_event_create_listener_thread (xine_data->xine_event_queue,
					   xine_event_cb, this);
	events_pending = 0;


	/*Tell xine to de-interlace all video-out data */
	xine_set_param (xine_data->xine_stream, XINE_PARAM_VO_DEINTERLACE,
			true);

	xine_data->osd_timeout = 3;
	xine_data->osd_enabled = 1;
	osd_init ();

	xine_data->mrl = NULL;

	return initialized = true;
}

void
xine_interface::deinit ()
{
	xine_dispose (xine_data->xine_stream);
	xine_close_audio_driver (xine_data->xine, xine_data->xine_audio);
	xine_close_video_driver (xine_data->xine, xine_data->xine_video);
	xine_data->completion_event = -1;
	initialized = 0;
}

void
xine_interface::set_ui_event (int event_id, char *event_data)
{
	events_pending = 1;
	ui_event_data = new char[strlen (event_data) + 1];
	strcpy (ui_event_data, event_data);
	ui_event_type = event_id;
}

bool
xine_interface::get_ui_event (int *event_id, char *event_data)
{
	if (events_pending)
	{
		*event_id = ui_event_type;
		strcpy (event_data, ui_event_data);
		events_pending = 0;
		return true;
	}
	return false;

}

const char *const *
xine_interface::get_video_drivers ()
{
	return xine_list_video_output_plugins (xine_data->xine);
}

char *
xine_interface::get_current_video_driver ()
{
	return xine_data->video_driver;
}

const char *const *
xine_interface::get_audio_drivers ()
{
	return xine_list_audio_output_plugins (xine_data->xine);
}

char *
xine_interface::get_current_audio_driver ()
{
	return xine_data->audio_driver;
}

const char *const *
xine_interface::get_post_plugins (int type)
{
	return xine_list_post_plugins_typed (xine_data->xine, type);
}

void
xine_interface::start_plugin (int id)
{
	int num_mrls = 0;

	char **plugin_mrl;	/*
				 * xine_get_autoplay_mrls (xine_data->xine,
				 * auto_plugins[id].
				 * plugin_id, &num_mrls); */
	plugin_mrl = get_plugin_autoplay_mrls (id);
	if (plugin_mrl != NULL)
	{
		open (plugin_mrl[0]);
		play ();
	}

}

char *
xine_interface::get_auto_plugin (int id)
{
	if ((id > -1) && (id < num_plugins + 1))
	{
		return auto_plugins[id].plugin_id;
	}
	else
		return NULL;
}

char *
xine_interface::get_auto_plugin_desc (int id)
{
	if ((id > -1) && (id < num_plugins + 1))
	{
		return auto_plugins[id].plugin_desc;
	}
	else
		return NULL;
}

char **
xine_interface::get_plugin_autoplay_mrls (int id)
{
	char **autoplay_mrls;

	int num_mrls = 0;
	if ((id > -1) && (id < num_plugins + 1))
	{
		autoplay_mrls =
			xine_get_autoplay_mrls (xine_data->xine,
						auto_plugins[id].plugin_id,
						&num_mrls);
		if (num_mrls > 0)
		{
			fprintf (stderr, "++++ There are %i mrls within %s\n",
				 num_mrls, autoplay_mrls[0]);
			return autoplay_mrls;
		}
	}

	return NULL;
}

char *
xine_interface::get_extensions ()
{
	return xine_get_file_extensions (xine_data->xine);
}

int
xine_interface::use_vis (bool enable, const char *vis_name)
{
	int was_playing, pos, time, length;
	xine_post_out_t *audio_source;
	if (status () == XINE_STATUS_PLAY)
	{
		was_playing = 1;
		get_stream_time (&pos, &time, &length);
	}

	if (enable
	    && ((xine_data->vis_name != vis_name) || !xine_data->vis_enabled))
	{
		stop ();
		fprintf (stderr, "Old vis: %s New vis: %s\n",
			 xine_data->vis_name, vis_name);
		xine_data->vis_name = (char *) vis_name;
		xine_data->vis_enabled = 1;
		xine_data->vis_active = 0;
		if (was_playing)
		{
			play ();
			set_position (pos);
		}

	}
	else if (xine_data->vis_enabled == 1 && !enable)
	{
		stop ();
		xine_data->vis_enabled = 0;
		if (was_playing)
		{
			set_position (pos);
		}
	}

	return xine_data->vis_enabled;

}

char *
xine_interface::get_vis ()
{
	if (xine_data->vis_enabled)
		return xine_data->vis_name;
	else
		return "none";
}

/***************************************************************************
** Media Playback and loading cotrols                                     **
***************************************************************************/

int
xine_interface::open (char *mrl)
{
	xine_data->mrl = mrl;
	return xine_open (xine_data->xine_stream, mrl);
}

bool
xine_interface::start_stream (int position)
{

	xine_post_out_t *audio_source =
		xine_get_audio_source (xine_data->xine_stream);
	
	if (xine_get_stream_info (xine_data->xine_stream,
				  XINE_STREAM_INFO_HAS_VIDEO)
	    && xine_get_stream_info (xine_data->xine_stream,
				     XINE_STREAM_INFO_VIDEO_HANDLED))
	{
		xine_post_wire_audio_port (audio_source,
					   xine_data->xine_audio);
		xine_data->vis_active = 0;
	}
	else if (xine_data->vis_enabled && !xine_data->vis_active)
	{
		xine_data->post_plugin = xine_post_init (xine_data->xine,
							 xine_data->vis_name, 0,
							 &xine_data->xine_audio,
							 &xine_data->xine_video);
		xine_post_wire_audio_port (audio_source,
					   xine_data->post_plugin->
					   audio_input[0]);
		xine_data->vis_active = 1;
	}
	else if (!xine_data->vis_enabled && xine_data->vis_active)
	{
		xine_post_wire_audio_port (audio_source,
					   xine_data->xine_audio);
		xine_data->vis_active = 0;
	}
	int result;
	if (xine_data->mrl)
		result = xine_play (xine_data->xine_stream, position, 0);
	return result;
}

bool
xine_interface::play ()
{
	if (status () != XINE_STATUS_PLAY)
		return start_stream (0);
	else if (get_speed () != 4)	//User wants to return to normal speed
		set_speed (4);
	else
		set_speed (0);	//User wants to pause the stream
	osd_update_status ();
	return 1;
}

void
xine_interface::stop ()
{
	xine_stop (xine_data->xine_stream);
	osd_update_status ();
	if (xine_data->vis_active)
	{		
		xine_post_out_t *audio_source =
		xine_get_audio_source (xine_data->xine_stream);
		xine_post_wire_audio_port (audio_source,
				   xine_data->xine_audio);
		xine_post_dispose(xine_data->xine,xine_data->post_plugin);
		xine_data->vis_active=0;
	}	
}

int
xine_interface::eject ()
{
	return xine_eject (xine_data->xine_stream);
}

void
xine_interface::faster ()
{
	int current_speed = get_speed ();
	if ((current_speed > 0) && (current_speed < 16))
	{
		set_speed (current_speed * 2);
	}
	else if (current_speed == 0)
	{
		set_speed (current_speed + 1);
	}
	osd_update_status ();
}

void
xine_interface::slower ()
{
	int current_speed = get_speed ();
	if ((current_speed > 1))
	{
		set_speed (current_speed / 2);
	}
	else if (current_speed == 1)
	{
		set_speed (current_speed - 1);
	}
	osd_update_status ();
}

void
xine_interface::set_speed (int new_speed)
{
	xine_set_param (xine_data->xine_stream, XINE_PARAM_SPEED, new_speed);
	osd_update_status ();
}

int
xine_interface::set_position (int jump_position)
{
	//int stream_length, temp_1, temp_2, result;
	//xine_get_pos_length (xine_data->xine_stream, &temp_1, &temp_2,
	//                   &stream_length);   
	//if ((jump_position < stream_length) && (jump_position > -1))
	//{
	fprintf (stderr, "Jumping to %i\n", jump_position);
	stop ();
	int result = start_stream (jump_position);
	//      xine_get_pos_length (xine_data->xine_stream, &temp_1, &temp_2,
	//                           &stream_length);
	osd_update_status ();
	//osd_stream_position (temp_2);
	return result;
	//}
	//else
	//      return false;
}

int
xine_interface::get_speed ()
{
	return xine_get_param (xine_data->xine_stream, XINE_PARAM_SPEED);
}

int
xine_interface::get_stream_time (int *stream_pos, int *position_time,
				 int *length_time)
{
	/*int stream_pos     Current relative position in the stream (0-65535)
	 * int position_time  Amount of time passed since stream start (millisecs)
	 * int length_time    Length of stream in millisecs */
	return xine_get_pos_length (xine_data->xine_stream, stream_pos,
				    position_time, length_time);
}

/*******************************************************************************
* Begin volume control                                                         *
*******************************************************************************/

int
xine_interface::get_volume ()
{
	return xine_get_param (xine_data->xine_stream,
			       XINE_PARAM_AUDIO_VOLUME);
}

void
xine_interface::set_volume (int volume)
{
	xine_set_param (xine_data->xine_stream, XINE_PARAM_AUDIO_VOLUME,
			volume);
	osd_draw_bar ("Volume", 0, 110, volume, OSD_BAR_STEPPER);
}

int
xine_interface::volume_up ()
{
	int cur_vol = get_volume ();
	if (cur_vol < 98)
		set_volume (cur_vol += 2);
	return cur_vol;
}

int
xine_interface::volume_down ()
{
	int cur_vol = get_volume ();
	if (cur_vol > 0)
		set_volume (cur_vol -= 2);
	return cur_vol;
}

/*******************************************************************************
* Begin navigation interfaces                                                  *
* these will be used for menu navigation by supported stream, such as a DVD    *
* menu, for instance.                                                          *
*******************************************************************************/
bool
xine_interface::next ()
{
	if (xine_get_stream_info
	    (xine_data->xine_stream, XINE_STREAM_INFO_HAS_CHAPTERS))
	{
		send_simple_event (XINE_EVENT_INPUT_NEXT);
		osd_update_status ();
		return true;
	}
	return false;
}

bool
xine_interface::prev ()
{
	if (xine_get_stream_info
	    (xine_data->xine_stream, XINE_STREAM_INFO_HAS_CHAPTERS))
	{
		send_simple_event (XINE_EVENT_INPUT_PREVIOUS);
		osd_update_status ();
		return true;
	}
	return false;
}

void
xine_interface::nav_up ()
{
	send_simple_event (XINE_EVENT_INPUT_UP);
}

void
xine_interface::nav_down ()
{
	send_simple_event (XINE_EVENT_INPUT_DOWN);
}

void
xine_interface::nav_left ()
{
	send_simple_event (XINE_EVENT_INPUT_LEFT);
}

void
xine_interface::nav_right ()
{
	send_simple_event (XINE_EVENT_INPUT_RIGHT);
}

void
xine_interface::nav_select ()
{
	send_simple_event (XINE_EVENT_INPUT_SELECT);
}

void
xine_interface::nav_menu1 ()
{
	send_simple_event (XINE_EVENT_INPUT_MENU1);
}

void
xine_interface::nav_menu2 ()
{
	send_simple_event (XINE_EVENT_INPUT_MENU2);
}

/*******************************************************************************
* Begin stream information                                                     *
* these functions can be called to retrieve information about the stream       *
*******************************************************************************/

int
xine_interface::status ()
{
	return xine_get_status (xine_data->xine_stream);
}

uint32_t
xine_interface::audio_bitrate ()
{
	return xine_get_stream_info (xine_data->xine_stream,
				     XINE_STREAM_INFO_AUDIO_BITRATE);
}

uint32_t
xine_interface::audio_samplerate ()
{
	return xine_get_stream_info (xine_data->xine_stream,
				     XINE_STREAM_INFO_AUDIO_SAMPLERATE);
}

bool
xine_interface::has_video ()
{
	return (bool) xine_get_stream_info (xine_data->xine_stream,
					    XINE_STREAM_INFO_HAS_VIDEO);
}

bool
xine_interface::has_audio ()
{
	return (bool) xine_get_stream_info (xine_data->xine_stream,
					    XINE_STREAM_INFO_HAS_AUDIO);
}

/*******************************************************************************
* Begin stream meta information                                                *
* these functions can be called to retrieve information within the stream, such*
* as MP3 ID3 tag															   *
* data.                                                                        *
*******************************************************************************/
const char *
xine_interface::mrl ()
{
	return xine_data->mrl;
}


const char *
xine_interface::title ()
{
	const char *title = xine_get_meta_info (xine_data->xine_stream,
						XINE_META_INFO_TITLE);
	if (title != NULL)
		return title;
	else
		return (const char *) xine_data->mrl;
}
const char *
xine_interface::artist ()
{
	return xine_get_meta_info (xine_data->xine_stream,
				   XINE_META_INFO_ARTIST);
}
const char *
xine_interface::album ()
{
	return xine_get_meta_info (xine_data->xine_stream,
				   XINE_META_INFO_ALBUM);
}
const char *
xine_interface::genre ()
{
	return xine_get_meta_info (xine_data->xine_stream,
				   XINE_META_INFO_GENRE);
}
const char *
xine_interface::comment ()
{
	return xine_get_meta_info (xine_data->xine_stream,
				   XINE_META_INFO_COMMENT);
}

/*******************************************************************************
* Begin Xine OSD functions                                                     *
* these functions can be called to display information over the video display  *
*******************************************************************************/

void
xine_interface::osd_init ()
{
	int fonth = 30;

	xine_data->osd_info =
		xine_osd_new (xine_data->xine_stream, 0, 0, 900,
			      (fonth * 6) + (5 * 3));
	xine_osd_set_font (xine_data->osd_info, "sans", fonth);
	xine_osd_set_text_palette (xine_data->osd_info,
				   XINE_TEXTPALETTE_WHITE_BLACK_TRANSPARENT,
				   XINE_OSD_TEXT1);
	const uint32_t *color;
	color = new uint32_t;
	color = (const uint32_t *) textpalettes_color;
	const uint8_t *trans;
	trans = new uint8_t;
	trans = textpalettes_trans;

	xine_data->osd_bar[0] =
		xine_osd_new (xine_data->xine_stream, 0, 0, BAR_WIDTH + 1,
			      BAR_HEIGHT + 1);
	//xine_osd_set_text_palette(xine_data->osd_bar[0], XINE_TEXTPALETTE_YELLOW_BLACK_TRANSPARENT, XINE_OSD_TEXT1);
	xine_osd_set_palette (xine_data->osd_bar[0], color, trans);

	xine_data->osd_bar[1] =
		xine_osd_new (xine_data->xine_stream, 0, 0, BAR_WIDTH + 1,
			      BAR_HEIGHT + 1);
	xine_osd_set_font (xine_data->osd_bar[1], "sans", fonth);
	xine_osd_set_text_palette (xine_data->osd_bar[1],
				   XINE_TEXTPALETTE_WHITE_BLACK_TRANSPARENT,
				   XINE_OSD_TEXT1);

	xine_data->osd_status =
		xine_osd_new (xine_data->xine_stream, 0, 0, 300, fonth + 4);
	xine_osd_set_font (xine_data->osd_status, "cetus", fonth);
	xine_osd_set_text_palette (xine_data->osd_status,
				   XINE_TEXTPALETTE_WHITE_BLACK_TRANSPARENT,
				   XINE_OSD_TEXT1);
	pthread_create (&xine_data->osd_timer_thread, NULL, osd_timer_thread,
			xine_data);

}

void
xine_interface::osd_deinit ()
{

	if (xine_data->osd_info_visible)
	{
		xine_data->osd_info_visible = 0;
		xine_osd_hide (xine_data->osd_info, 0);
	}

	if (xine_data->osd_bar_visible)
	{
		xine_data->osd_bar_visible = 0;
		xine_osd_hide (xine_data->osd_bar[0], 0);
		xine_osd_hide (xine_data->osd_bar[1], 0);
	}

	if (xine_data->osd_status_visible)
	{
		xine_data->osd_status_visible = 0;
		xine_osd_hide (xine_data->osd_status, 0);
	}

	xine_osd_free (xine_data->osd_info);
	xine_osd_free (xine_data->osd_bar[0]);
	xine_osd_free (xine_data->osd_bar[1]);
	xine_osd_free (xine_data->osd_status);
}

void
xine_interface::osd_update ()
{


}

void
xine_interface::osd_stream_infos (void)
{

	if (xine_data->osd_enabled)
	{
		uint32_t vwidth, vheight, asrate;
		const char *vcodec, *acodec;
		char buffer[256], *p;
		int x, y;
		int w, h, osdw;
		int playedtime, totaltime, pos;
		int audiochannel, spuchannel, len;

		vcodec = xine_get_meta_info (xine_data->xine_stream,
					     XINE_META_INFO_VIDEOCODEC);
		acodec = xine_get_meta_info (xine_data->xine_stream,
					     XINE_META_INFO_AUDIOCODEC);
		vwidth = xine_get_stream_info (xine_data->xine_stream,
					       XINE_STREAM_INFO_VIDEO_WIDTH);
		vheight =
			xine_get_stream_info (xine_data->xine_stream,
					      XINE_STREAM_INFO_VIDEO_HEIGHT);
		asrate = xine_get_stream_info (xine_data->xine_stream,
					       XINE_STREAM_INFO_AUDIO_SAMPLERATE);
		audiochannel =
			xine_get_param (xine_data->xine_stream,
					XINE_PARAM_AUDIO_CHANNEL_LOGICAL);
		spuchannel =
			xine_get_param (xine_data->xine_stream,
					XINE_PARAM_SPU_CHANNEL);

		if (!xine_get_pos_length
		    (xine_data->xine_stream, &pos, &playedtime, &totaltime))
			return;

		playedtime /= 1000;
		totaltime /= 1000;

		xine_osd_clear (xine_data->osd_info);

		/* We're in visual animation mode */
		if ((vwidth == 0) && (vheight == 0))
		{
			//if(xine_data->vis_enabled == 1)
			//{
			//video_window_get_frame_size(&vwidth, &vheight);
			//vwidth = xine_data->video_attr[4];
			//vheight = xine_data->video_attr[5];
			vcodec = xine_data->vis_name;
			//}
			if (xine_data->vis_enabled == 1)
			{
				vcodec = xine_get_meta_info (xine_data->
							     xine_stream,
							     XINE_META_INFO_VIDEOCODEC);
				vwidth = xine_get_stream_info (xine_data->
							       xine_stream,
							       XINE_STREAM_INFO_VIDEO_WIDTH);
				vheight =
					xine_get_stream_info (xine_data->
							      xine_stream,
							      XINE_STREAM_INFO_VIDEO_HEIGHT);
			}
			else
			{
				vwidth = xine_data->video_attr[4];
				vheight = xine_data->video_attr[5];
				vcodec = "unknown";
			}
		}

		y = x = 0;

		sprintf (buffer, "%s", xine_data->mrl);
//    sprintf(buffer,"Test MRL fixme");
		xine_osd_get_text_size (xine_data->osd_info, buffer, &osdw,
					&h);
		p = buffer;
		while (osdw > (vwidth - 40))
		{
			*(p++) = '\0';
			*(p) = '.';
			*(p + 1) = '.';
			*(p + 2) = '.';
			xine_osd_get_text_size (xine_data->osd_info, p, &osdw,
						&h);
		}
		xine_osd_draw_text (xine_data->osd_info, x, y, p,
				    XINE_OSD_TEXT1);

		y += h;

		if (vcodec && vwidth && vheight)
		{
			sprintf (buffer, "%s: %dX%d", vcodec, vwidth,
				 vheight);
			xine_osd_draw_text (xine_data->osd_info, x, y, buffer,
					    XINE_OSD_TEXT1);
			xine_osd_get_text_size (xine_data->osd_info, buffer,
						&w, &h);
			if (w > osdw)
				osdw = w;
			y += h;
		}

		if (acodec && asrate)
		{
			sprintf (buffer, "%s: %dHz", acodec, asrate);
			xine_osd_draw_text (xine_data->osd_info, x, y, buffer,
					    XINE_OSD_TEXT1);
			xine_osd_get_text_size (xine_data->osd_info, buffer,
						&w, &h);
			if (w > osdw)
				osdw = w;
			y += h;
		}

		memset (&buffer, 0, sizeof (buffer));

		sprintf (buffer, "Audio: ");
		len = strlen (buffer);
		switch (audiochannel)
		{
		case -2:
			sprintf (buffer, "%soff", buffer);
			break;
		case -1:
			if (!xine_get_audio_lang
			    (xine_data->xine_stream, audiochannel,
			     &buffer[len]))
				sprintf (buffer, "%sauto", buffer);
			break;
		default:
			if (!xine_get_audio_lang
			    (xine_data->xine_stream, audiochannel,
			     &buffer[len]))
				sprintf (buffer, "%s%3d", buffer,
					 audiochannel);
			break;
		}

		sprintf (buffer, "%s, Spu: ", buffer);
		len = strlen (buffer);
		switch (spuchannel)
		{
		case -2:
			sprintf (buffer, "%soff", buffer);
			break;
		case -1:
			if (!xine_get_spu_lang
			    (xine_data->xine_stream, spuchannel,
			     &buffer[len]))
				sprintf (buffer, "%sauto", buffer);
			break;
		default:
			if (!xine_get_spu_lang
			    (xine_data->xine_stream, spuchannel,
			     &buffer[len]))
				sprintf (buffer, "%s%3d", buffer, spuchannel);
			break;
		}
		sprintf (buffer, "%s.", buffer);
		xine_osd_draw_text (xine_data->osd_info, x, y, buffer,
				    XINE_OSD_TEXT1);
		xine_osd_get_text_size (xine_data->osd_info, buffer, &w, &h);
		if (w > osdw)
			osdw = w;

		y += (h);

		if (totaltime)
		{
			sprintf (buffer,
				 "%d:%.2d:%.2d (%.0f%%) of %d:%.2d:%.2d",
				 playedtime / 3600, (playedtime % 3600) / 60,
				 playedtime % 60,
				 ((float) playedtime / (float) totaltime) *
				 100, totaltime / 3600,
				 (totaltime % 3600) / 60, totaltime % 60);
			xine_osd_draw_text (xine_data->osd_info, x, y, buffer,
					    XINE_OSD_TEXT1);
			xine_osd_get_text_size (xine_data->osd_info, buffer,
						&w, &h);
			if (w > osdw)
				osdw = w;

			osd_stream_position (pos);
		}

		x = (vwidth - osdw) - 40;
		xine_osd_set_position (xine_data->osd_info, (x >= 0) ? x : 0,
				       15);

		xine_osd_show (xine_data->osd_info, 0);
		xine_data->osd_info_visible = xine_data->osd_timeout;
	}
}

void
xine_interface::osd_draw_bar (char *title, int min, int max, int val,
			      int type)
{

	if (xine_data->osd_enabled)
	{
		uint32_t vwidth, vheight;
		int bar_color[40];
		int i, x;
		float _val = (int) val;
		float _min = (int) min;
		float _max = (int) max;
		int pos;

		if (max <= min)
			_max = (int) (min + 1);
		if (min >= max)
			_min = (int) (max - 1);
		if (val > max)
			_val = (int) max;
		if (val < min)
			_val = (int) min;

		pos = (int) (_val + -_min) / (int) ((_max + -_min) / 40);

		xine_osd_clear (xine_data->osd_bar[0]);
		xine_osd_clear (xine_data->osd_bar[1]);

		memset (&bar_color, (XINE_OSD_TEXT1 + 7), sizeof (int) * 40);

		switch (type)
		{
		case OSD_BAR_PROGRESS:
		case OSD_BAR_STEPPER:
			if (pos)
				memset (bar_color, (XINE_OSD_TEXT1 + 21),
					sizeof (int) * pos);
			break;
		case OSD_BAR_POS:
			if (pos)
				bar_color[pos - 1] = (XINE_OSD_TEXT1 + 21);
			break;
		}
		if ((type == OSD_BAR_PROGRESS) || (type == OSD_BAR_POS))
		{
			x = 3;
			xine_osd_draw_rect (xine_data->osd_bar[0], x, 2,
					    x + 3, BAR_HEIGHT - 2,
					    XINE_OSD_TEXT1 + 9, 1);
			x += 8;

			for (i = 0; i < 40; i++, x += 8)
			{
				xine_osd_draw_rect (xine_data->osd_bar[0],
						    x, 6, x + 3,
						    BAR_HEIGHT - 2,
						    bar_color[i], 1);
			}

			xine_osd_draw_rect (xine_data->osd_bar[0],
					    x, 2, x + 3, BAR_HEIGHT - 2,
					    XINE_OSD_TEXT1 + 9, 1);
		}
		else if (type == OSD_BAR_STEPPER)
		{
			int y = BAR_HEIGHT - 4;
			int step = y / 20;

			x = 11;

			for (i = 0; i < 40; i++, x += 8)
			{
				xine_osd_draw_rect (xine_data->osd_bar[0],
						    x, y, x + 3,
						    BAR_HEIGHT - 2,
						    bar_color[i], 1);

				if (!(i % 2))
					y -= step;
			}
		}

		if (title)
		{
			int tw, th;

			xine_osd_get_text_size (xine_data->osd_bar[1], title,
						&tw, &th);
			xine_osd_draw_text (xine_data->osd_bar[1],
					    (BAR_WIDTH - tw) >> 1, 0, title,
					    XINE_OSD_TEXT1);
		}

		vwidth = xine_get_stream_info (xine_data->xine_stream,
					       XINE_STREAM_INFO_VIDEO_WIDTH);
		vheight =
			xine_get_stream_info (xine_data->xine_stream,
					      XINE_STREAM_INFO_VIDEO_HEIGHT);

		/* We're in visual animation mode */
		if ((vwidth == 0) && (vheight == 0))
		{
			if (xine_data->vis_enabled == 1)
			{
				vwidth = xine_data->video_attr[4];
				vheight = xine_data->video_attr[5];
			}
			else if (xine_data->vis_enabled == 2)
			{
				vwidth = xine_get_stream_info (xine_data->
							       xine_stream,
							       XINE_STREAM_INFO_VIDEO_WIDTH);
				vheight =
					xine_get_stream_info (xine_data->
							      xine_stream,
							      XINE_STREAM_INFO_VIDEO_HEIGHT);
			}
		}

		x = (vwidth - BAR_WIDTH) >> 1;
		xine_osd_set_position (xine_data->osd_bar[0],
				       (x >= 0) ? x : 0,
				       (vheight - BAR_HEIGHT) - 40);
		xine_osd_set_position (xine_data->osd_bar[1],
				       (x >= 0) ? x : 0,
				       (vheight - (BAR_HEIGHT * 2)) - 40);

		xine_osd_show (xine_data->osd_bar[0], 0);
		if (title)
			xine_osd_show (xine_data->osd_bar[1], 0);

		xine_data->osd_bar_visible = xine_data->osd_timeout;
	}
}

void
xine_interface::osd_stream_position (int pos)
{
	osd_draw_bar ("Position in Stream", 0, 65535, pos, OSD_BAR_PROGRESS);
}

void
xine_interface::osd_update_status ()
{

	if (xine_data->osd_enabled)
	{
		int status;
		char buffer[256];

		status = xine_get_status (xine_data->xine_stream);

		xine_osd_clear (xine_data->osd_status);

		/*
		 * { : eject
		 * [ : previous
		 * | : | (thin)
		 * @ : | (large)
		 * ] : next
		 * } : stop
		 * $ : > (large)
		 * > : play
		 * < : pause
		 */

		memset (&buffer, 0, sizeof (buffer));

		switch (status)
		{
		case XINE_STATUS_IDLE:
		case XINE_STATUS_STOP:
			sprintf (buffer, "%s",
				 (_osd_get_status_sym (status)));
			break;

		case XINE_STATUS_PLAY:
		{
			int speed = xine_get_param (xine_data->xine_stream,
						    XINE_PARAM_SPEED);
			int secs;

			if (xine_get_pos_length
			    (xine_data->xine_stream, NULL, &secs, NULL))
			{
				secs /= 1000;
				sprintf (buffer, "%s %02d:%02d:%02d",
					 (_osd_get_speed_sym (speed)),
					 secs / (60 * 60), (secs / 60) % 60,
					 secs % 60);
			}
			else
				sprintf (buffer, "%s",
					 (_osd_get_speed_sym (speed)));
		}
			break;
		case XINE_STATUS_QUIT:
			/* noop */
			break;
		}

		xine_osd_draw_text (xine_data->osd_status, 0, 0, buffer,
				    XINE_OSD_TEXT1);
		xine_osd_set_position (xine_data->osd_status, 20, 10);
		xine_osd_show (xine_data->osd_status, 0);
		xine_data->osd_status_visible = xine_data->osd_timeout;
	}
}


const char *
xine_interface::config_new_string (const char *key, const char *def_value,
				   const char *description, const char *help,
				   int exp_level, xine_config_cb_t changed_cb,
				   void *cb_data)
{
	return xine_config_register_string (xine_data->xine, key, def_value,
					    description, help, exp_level,
					    changed_cb, cb_data);
}

int
xine_interface::config_new_range (const char *key, int def_value,
				   int min, int max, const char *description,
				   const char *help, int exp_level,
				   xine_config_cb_t changed_cb, void *cb_data)
{
	return xine_config_register_range (xine_data->xine, key, def_value,
					   min, max, description, help,
					   exp_level, changed_cb, cb_data);
}

int
xine_interface::config_new_bool (const char *key, int def_value,
					const char *description,
				   const char *help, int exp_level,
				   xine_config_cb_t changed_cb, void *cb_data)
{
	return xine_config_register_bool (xine_data->xine, key, def_value,
					   description, help,
					   exp_level, changed_cb, cb_data);
}


int
xine_interface::config_lookup (const char *key, xine_cfg_entry_t * entry)
{
	return xine_config_lookup_entry (xine_data->xine, key, entry);
}

void
xine_interface::config_update (xine_cfg_entry_t * entry)
{
	xine_config_update_entry (xine_data->xine, entry);
}

void
xine_interface::config_save ()
{
	fprintf (stdout, "Storing configuration settings in %s....",
		 config_file);
	xine_config_save (xine_data->xine, config_file);
	fprintf (stdout, "Done\n");
}
